import pyspark.sql.functions as F
from pyspark.sql.utils import AnalysisException
from pyspark.sql.functions import expr

class Audit():

    def __init__(self,
                 spark_session,
                 master_location,
                 project_config,
                 _log):

        self.log = _log
        self.spark = spark_session
        self.master_location = master_location
        self.project_config = project_config

    def unpivot(self, df):

        unpivotExpr = """stack(22,   
            'avg hcc captured', avg_hcc_captured,  
            'avg hcc opportunity', avg_hcc_opp,  
            'avg raf score', avg_raf,  
            'avgdemo', avg_raf_demo,  
            'avg projected raf score', avg_raf_projected,  
            'eligible members', cast(eligible_count as double),  
            'female population ratios', cast(female_percent as double),  
            'members w/ zero hcc score', cast(hcc_zero_count as double),  
            'male population ratios', cast(male_percent as double),  
            'members w/ no provider visits', cast(no_dos_count as double),  
            'no_wellness_visit_count', cast(no_wellness_visit_count as double),  
            'one_dos_and_hcc_sus_count', cast(one_dos_and_hcc_sus_count as double),  
            'gtr_than_three_sus_count', cast(gtr_than_three_sus_count as double),  
            'members w/ no pcp visits', cast(no_pcp_dos_count as double),  
            'members w/ zero demo score', cast(no_demo_count as double),  
            'recapture rates', cast(recapture_rate as double),  
            'RAFx call Errors', cast(rafx_call_errors as double),  
            'total_captured_dxs_sent_to_lambda_rafx', cast(total_captured_dxs_sent_to_lambda_rafx as double),  
            'total_suspect_dxs_sent_to_lambda_rafx', cast(total_suspect_dxs_sent_to_lambda_rafx as double),  
            'total_input_dxs', cast(total_input_dxs as double),  
            'total_hccs', cast(total_hccs as double),  
            'Number of members without first name', cast(first_name_lesseq_than_1_count as double)  
            )  as ( measure, value) 
            """

        df.select(
            "record_timestamp", "job_id", "year",
            "year_name", expr(unpivotExpr)
        ).where("value is not null ").createOrReplaceTempView("unpivot_df")

        return self.spark.sql("""
                select *,
                row_number() over(partition by measure, year     order by  record_timestamp desc) as r ,
                case
                when measure = 'avg hcc captured' then '4kpi'
                when measure = 'avg hcc opportunity' then 'dashboard10'
                when measure = 'avg raf score' then '1kpi'
                when measure = 'avgdemo' then 'dashboard06'
                when measure = 'avg projected raf score' then 'dashboard08'
                when measure = 'eligible members' then '3kpi'
                when measure = 'female population ratios' then 'dashboard12'
                when measure = 'members w/ zero hcc score' then 'dashboard02'
                when measure = 'male population ratios' then 'dashboard11'
                when measure = 'members w/ no provider visits' then 'dashboard04'
                when measure = 'recapture rates' then '2kpi'
                when measure = 'RAFx call Errors' then '5kpi'
                when measure = 'members w/ no pcp visits' then 'dashboard03'
                when measure = 'no_demo_count' then 'dashboard05'
                when measure = 'Number of members without first name' then '6kpi'
                else 'OTHER_GROUP' end as groupy
                from unpivot_df
                where year between 2019 and  year(current_date())
            """)





"""
    This is a template for a tear down job. The teardown job removes
    mnt/vol/local.yaml file. This is the last step for dag.
"""

import yaml
import os
import argparse
import json
from dependencies.logger import logger
from dependencies.post_gres import CadmiumPostGres


class TearDown:
    def __init__(self, _config=None):
        self.vault_metadata = {}
        self.postgres = {}
        self.client = "THP"
        self.user = "USER1"
        self.workflow = "SSD"
        self.env = "test"
        self.table_name = "job"
        self._log = logger()

    def init_config(self, _config):
        self.vault_metadata = _config.get('vault_metadata', self.vault_metadata)
        self.postgres = _config.get('job_history').get('postgres', self.postgres)
        self.client = _config.get('client', self.client)
        self.user = _config.get('user', self.user)
        self.workflow = _config.get('workflow', self.workflow)
        self.env = _config.get('env', self.env)

    def clean_vault_metadata(self):
        a = self.vault_metadata.keys()
        selection = [1, 2, 3]
        for el in a:
            if el in selection:
                self.vault_metadata[el].pop('deletion_time', None)
                self.vault_metadata[el].pop('destroyed', None)

    def get_config(self):
        """
            This function retrieves argument from DAG yaml template.
          :return:
          """

        parser = argparse.ArgumentParser(
            description="Argument Parser for teardown "
        )
        parser.add_argument(
            "--mnt_volume", action="store", type=str, required=False, help="mount volume path ",
        )
        args = vars(parser.parse_args())
        config = {"mnt_volume": args["mnt_volume"]}
        return config

    @property
    def log(self):
        return self._log


if __name__ == "__main__":
    # Read and delete
    _local_path = '/mnt/vol/local.yaml'
    td = TearDown()
    _mnt_config = td.get_config()
    local_path = _mnt_config.get('mnt_volume', _local_path)  # from argument

    try:
        with open(local_path, "r") as conf_file:
            __config = yaml.safe_load(conf_file)
        td.init_config(__config)
        _pg_values = td.postgres
        # td.log.info(__config)
        td.clean_vault_metadata()
        _vault_metadata = td.vault_metadata
        # td.log.info(_vault_metadata)
        table_name = _pg_values.get('table_name', td.table_name)
        post_gres = CadmiumPostGres(_pg_values['hostname'], _pg_values['database_name'], _pg_values['user'],
                                    _pg_values['password'], _pg_values['port'], True)

        post_gres.insert_into_table(table_name=table_name,
                                    columns_tuples='(client_name,workflow_name,stage_name,user_name,vault_versions)',
                                    placeholders_tuple='(%s, %s,%s, %s,%s)',
                                    values_tuple=(td.client, td.workflow, td.env, td.user, json.dumps(_vault_metadata)))
        post_gres.close()
        td.log.info("=====================Removing local file =====================")
        os.remove(local_path)  # remove mount file
        td.log.info("Completed")

    except FileNotFoundError:
        td.log.error(f"Sorry, the file  does not exist, {local_path}")



"""
test_vault.py
~~~~~~~~~~~~~~~
This module contains unit tests for vault functionality tests. Test assertions are to
ensure payload data format are within expectations. As version number for the vault
payload is important, a few assertions are made to ensure it exists.

Another feature the vault request test must have is the ability to retry when the url
request is not available. Currently, for test purposes the retry is set to 2.

Assumptions:
Since, pytest is run locally or at the Travis CI environment, a direct connection to
the EpiSource Hashicorp Vault is not available as those environments, i.e. local and
Travis CI, are not within the same AWS Security Group, thus assertions are tested for
functionality tests.

"""

import unittest
import requests
import requests_mock
import time


class VaultTest:
    def __init__(self):
        self.data = self.data_on_success()
        self.job_history = {
            "job_run_at": "2020-07-09 11:12:00",
            "create_at": "2020-07-09 11:12:00",
            1: {
                "created_time": "2020-07-07T15:34:01.598860826Z",
                "deletion_time": "",
                "destroyed": False,
                "url_path": "conf/cadmium",
                "version": 1
            },
            2: {
                "created_time": "2020-07-07T15:34:01.598860826Z",
                "deletion_time": "",
                "destroyed": False,
                "url_path": "conf/thp",
                "version": 1
            },
            3: {
                "created_time": "2020-07-07T15:34:01.598860826Z",
                "deletion_time": "",
                "destroyed": False,
                "url_path": "infra/thp",
                "version": 1
            }
        }

    @staticmethod
    def data_on_success():
        cadmium_pl = {
            "request_id": "e8602044-0fdb-6290-234a-e185a2849a0c",
            "lease_id": "",
            "renewable": False,
            "lease_duration": 0,
            "data": {"data": {"value": "name: CADMIUM\npath: conf/cadmium\n"},
                     "metadata": {"created_time": "2020-07-07T15:34:01.598860826Z",
                                  "deletion_time": "", "destroyed": False, "version": 2}},
            "wrap_info": None, "warnings": None, "auth": None
        }

        with requests_mock.Mocker() as mock_request:
            mock_request.get("http://127.0.0.1:8200/v1/secret/data/cadmium/conf/cadmium", json=cadmium_pl)
            response = requests.get("http://127.0.0.1:8200/v1/secret/data/cadmium/conf/cadmium")

        return response.json()

    def vault_secrets_with_retry(self, vault_token, url, retry=5, _initial_retry=5, _url_count=0, _timeout=0.10):
        rj = self.read_from_vault(vault_token, url, _url_count=_url_count, _timeout=_timeout)
        msg = f"Request timeout at {_timeout}"
        if rj != msg:
            completed = _initial_retry - retry
            print("Success getting vault secrets , {}".format(completed))
            return msg
        if retry == 0:
            msg1 = "Something is wrong here at vault {}".format(url)
            return msg1
        if retry > 0:
            time.sleep(1)
            retry = retry - 1
            print("retry")
            print(retry)
            self.vault_secrets_with_retry(vault_token, url, retry=retry, _initial_retry=_initial_retry,
                                          _url_count=_url_count)

    def get_last_nth(self, _url, n):
        if _url:
            return "/".join(_url.split("/")[-n:])
        return ""

    def read_from_vault(self, vault_token, url, _url_count, _timeout):
        _job_history = self.job_history
        payload = {}
        try:
            payload = requests.get(
                url,
                headers={'X-Vault-Token': vault_token},
                timeout=_timeout
            )
        except requests.exceptions.Timeout:
            return f"Request timeout at {_timeout}"
        except requests.exceptions.TooManyRedirects:
            print("Too many redirect")
        except requests.exceptions.RequestException as e:  # This is the correct syntax
            print("there is an error", e)
            return

        __code = payload.get('status_code', "404")
        if __code == 200:
            pl_json = payload.json()
            # With version
            value_version = pl_json['data'].get('data', '')
            if value_version:
                value = pl_json['data']['data']['value']  # with version
                vault_metadata = pl_json['data']['metadata']
                vault_metadata["url_path"] = self.get_last_nth(url, 2)
                self.job_history[_url_count] = vault_metadata
                return value

            # No versioning.
            value = pl_json['data']['value']  # no version
            return value

        return __code


class TestHTTPRequest(unittest.TestCase):

    def test_request_vault_retry(self):
        _timeout = 0.10
        data = VaultTest().vault_secrets_with_retry("aTokenValue",
                                                    "http://10.1.1.1:8200/v1/secret/cadmium/conf/cadmium",
                                                    retry=2, _initial_retry=2,
                                                    _url_count=1,
                                                    _timeout=_timeout)

        assert data is None

    def test_request_vault_timeout(self):
        _timeout = 2.18
        data = VaultTest().read_from_vault("aTokenValue", "http://10.1.1.1:8200/v1/secret/cadmium/conf/cadmium", 1,
                                           _timeout=_timeout)
        assert data == f"Request timeout at {_timeout}"

    def test_lease_duration(self):
        d = VaultTest().data
        assert d["lease_duration"] == 0
        assert "request_id" in d.keys()

    def test_data_path(self):
        d = VaultTest().data
        assert d["data"]["data"]["value"] == "name: CADMIUM\npath: conf/cadmium\n"

    def test_vault_with_version(self):
        d = VaultTest().data
        assert "version" in d["data"]["metadata"].keys()

    def test_vault_with_version_number(self):
        d = VaultTest().data
        assert d["data"]["metadata"]["version"] == 2



"""
conftest.py
~~~~~~~~~~~~~~~

https://docs.pytest.org/en/latest/parametrize.html#pytest-generate-tests
"""
import pytest
import pandas as pd
from tests.spark_context_fixture import spark_context, spark_session
from dependencies import aws, config

pytestmark = pytest.mark.usefixtures("spark_session")
CONFIG_PROJECT_TEST_FILE_NAME = 'project_test.yaml'


def get_test_etl_job_parameters():
    '''

    :return:
    '''
    project_configs =  config.load_framework_config()[CONFIG_PROJECT_TEST_FILE_NAME]
    return project_configs['test_etl_job']['jobs']


def pytest_generate_tests(metafunc):
    '''

    :param metafunc:
    :return:
    '''
    if "etl_job_test_modules" in metafunc.fixturenames:
        metafunc.parametrize("etl_job_test_module", get_test_etl_job_parameters())



import re
from dependencies.exception import exception_handler
from jobs.etl_job_base import EtlJobBase
import pyspark.sql.functions as F
import pyspark.sql.types as T
from pyspark.sql.utils import AnalysisException
from constants.constants_configs import Config as Constant_Config

# from dependencies.unit_test_data_factory import unit_test_data_factory


class JSONToParquet(EtlJobBase):
    @exception_handler
    def get_custom_argparse_configs(self, parser):
        """
        :params parser: the arg parse object to add custom CLI params to
        :return parser:
        """
        parser.add_argument(
            "--input-source",
            action="store",
            type=str,
            required=False,
            help="if input source need be specified",
        )
        parser.add_argument(
            "--output-loop-location",
            action="store",
            type=str,
            required=True,
            help="S3 location where to write outputloop files.  Suffix of the output SSD table name will be appended to this path",
        )
        return parser

    @exception_handler
    def add_custom_config_fields(self, config, args):
        """
        Overloading to add custom config fields
        :params config: config object created in get_base_argparse_configs
        :params args: args given to argparse, used in creating custom config pieces
        :return config: new config object with desired custom fields
        """

        config[Constant_Config.STEP_CONFIG]["input_source"] = args["input_source"]
        config[Constant_Config.STEP_CONFIG]["output_loop_location"] = args[
            "output_loop_location"
        ]
        config[Constant_Config.STEP_CONFIG]["metadata"] = {}

        if "837" in config["step_name"]:
            config[Constant_Config.STEP_CONFIG]["file_type"] = "837"
        elif "834" in config["step_name"]:
            config[Constant_Config.STEP_CONFIG]["file_type"] = "834"
        elif "rx" in config["step_name"]:
            config[Constant_Config.STEP_CONFIG]["file_type"] = "rx"

        return config

    @exception_handler
    def create_output_loop(self, df, output_loop):
        """
        Create SSD tables based on output_loop specified in YAML
        :params df: SSD dataframe created by outputloop
        :params output_loop: yaml config from step yaml file for SSD output loop
        :returns df: SSD dataframe
        """
        # TODO: in create_output_loop, check to make sure each field is part of the schema. drop or error if not.
        df_loop = df
        self.log.info(f"WORKING ON {output_loop['name']} OUTPUT LOOP")
        self.config[Constant_Config.STEP_CONFIG]["metadata"][output_loop["name"]] = []
        if output_loop["list_key"] is not None:
            # Renaming doesnt work with nested structs. If i keep a top level struct with the same name, nested references will not refer to the exploded column.
            df_loop = df_loop.withColumn(
                output_loop["list_key"].split(".")[-1] + "_exp",
                F.explode(df_loop[output_loop["list_key"]]),
            )
        for field in output_loop["fields"].keys():
            field_info = output_loop["fields"][field]
            if output_loop["list_key"] is not None:
                field_info["src_field"] = field_info["src_field"].replace(
                    output_loop["list_key"],
                    output_loop["list_key"].split(".")[-1] + "_exp",
                )
            # Check if field refers to metadata, to be added later
            if re.fullmatch(r".*\{.*\}.*", field_info["src_field"]):
                self.config[Constant_Config.STEP_CONFIG]["metadata"][
                    output_loop["name"]
                ].append(field)
                df_loop = df_loop.withColumn(field, F.lit(None))
            else:
                if "alt_src_field" in field_info.keys():
                    field_info["alt_src_field"] = field_info["alt_src_field"].replace(
                        output_loop["list_key"],
                        output_loop["list_key"].split(".")[-1] + "_exp",
                    )
                    # TODO: add try except for if src or alt_src are in schema
                    df_loop = df_loop.withColumn(
                        field,
                        F.coalesce(
                            df_loop[field_info["src_field"]],
                            df_loop[field_info["alt_src_field"]],
                        ),
                    )
                else:
                    try:
                        if "src_type" in field_info:
                            df_loop = df_loop.withColumn(
                                field, df_loop[field_info["src_field"]].cast(field_info['src_type'])
                            )
                        else:
                            df_loop = df_loop.withColumn(
                                field, df_loop[field_info["src_field"]]
                            )
                    except Exception:
                        self.log.warn(
                            f"{self.config[Constant_Config.STEP_CONFIG]['file_type']} from {self.config[Constant_Config.STEP_CONFIG]['input_source']} does not have field {field_info['src_field']}; set as Null"
                        )
                        print(field_info)
                        src_type = field_info['src_type'] if 'src_type' in field_info else 'string'
                        src_default = field_info['src_default'] if 'src_default' in field_info else None
                        df_loop = df_loop.withColumn(field,  F.lit(None if src_default=='null' else src_default).cast(src_type))
                        pass
        return df_loop.select(list(output_loop["fields"].keys()))

    @exception_handler
    def collapse_stupid_lists(self, df, output_loop):
        """
        Several of the fields in the json doc are the same but with different suffixes, _01..._12.
        We want to collapse all of these into a single field
        :params df: SSD dataframe created by outputloop
        :params output_loop: yaml config from step yaml file for SSD output loop
        :returns df: SSD dataframe with stupid lists collapsed
        """
        # Get All the stupid list columns; map to output column name
        stupid_list_cols = {}
        for k, v in output_loop["fields"].items():
            if re.search(r"_\d\d$", v["src_field"]):
                stupid_list_cols[v["src_field"]] = k
            if "alt_src_field" in v.keys():
                if re.search(r"_\d\d$", v["alt_src_field"]):
                    stupid_list_cols[v["alt_src_field"]] = k
        # Create Column Sets (columns that will go together)
        stupid_column_sets = {}
        for stupid_in_col, stupid_out_col in stupid_list_cols.items():
            try:
                assert re.search(r"_\d\d$", stupid_out_col)
            except AssertionError:
                self.log.warn(
                    f"{stupid_out_col} should have same suffix pattern as {stupid_in_col}"
                )
            if (
                re.split(r"_\d\d$", stupid_out_col, maxsplit=1)[0]
                in stupid_column_sets.keys()
            ):
                stupid_column_sets[
                    re.split(r"_\d\d$", stupid_out_col, maxsplit=1)[0]
                ].append(stupid_out_col)
            else:
                stupid_column_sets[
                    re.split(r"_\d\d$", stupid_out_col, maxsplit=1)[0]
                ] = [stupid_out_col]

        # Aggregate stupid list columns and delete originals
        for stupid_col_root, stupid_out_cols in stupid_column_sets.items():
            for stupid_out_col in list(stupid_out_cols):
                if df.schema[stupid_out_col].dataType == T.StringType():
                    df = df.withColumn(
                        stupid_out_col,
                        F.split(F.col(stupid_out_col), ",").cast(
                            T.ArrayType(T.StringType())
                        ),
                    )
                df = df.withColumn(
                    stupid_out_col,
                    F.coalesce(F.col(stupid_out_col), F.array().cast("array<string>")),
                )
            df = df.withColumn(
                stupid_col_root, F.array_remove(F.flatten(F.array(stupid_out_cols)), "")
            )
            df = df.drop(*stupid_out_cols)

        return df

    def add_metadata(self, df, output_loop):
        """
        Add metadata fields to SSD dataframe according to whether metadata field present in output-loop yaml
        :params df: SSD dataframe created by outputloop
        :params output_loop: yaml config from step yaml file for SSD output loop
        :returns df: SSD dataframe with metadata added
        """
        # Metadata columns: source, job_id, created_at
        metadata_fields = list(
            self.config[Constant_Config.STEP_CONFIG]["metadata"][output_loop["name"]]
        )
        if "cad_source_name" in metadata_fields:
            df = df.withColumn(
                "cad_source_name",
                F.lit(self.config[Constant_Config.STEP_CONFIG]["input_source"]),
            )
        if "cad_client" in metadata_fields:
            df = df.withColumn("cad_client", F.lit(self.config["client"]))
        if "cad_created_at" in metadata_fields:
            df = df.withColumn("cad_created_at", F.lit(F.current_timestamp()))
        if "cad_job_id" in metadata_fields:
            df = df.withColumn("cad_job_id", F.lit(self.config["job_id"]))
        return df

    @exception_handler
    def extract_data(self):
        """Load test_data from Parquet file format.
        :param spark: Spark session object.
        :return: Spark DataFrame.
        """
        # self.log.info(self.config[Constant_Config.STEP_CONFIG]["input_source"])
        df = self.spark.read.format("json").option('dropFieldIfAllNull', True).load(
            self.config[Constant_Config.STEP_CONFIG]["input_source"]
        )
        self.log.info(
            f"Writing json data from {self.config[Constant_Config.STEP_CONFIG]['input_source']}"
        )
        return df

    @exception_handler
    # @unit_test_data_factory
    def transform_data(self, df, output_loop):
        """Transform original dataset.
        :param df: Input DataFrame.
        :return: Transformed DataFrame.
        """
        # Run check on loop_key existence - if no list key, do not load ssd table.
        if output_loop["list_key"] is not None:
            try:
                df.select(output_loop["list_key"])
            except AnalysisException:
                # return
                self.log.warn(
                    f"List key {output_loop['list_key']} does not exist in {self.config[Constant_Config.STEP_CONFIG]['input_source']}"
                )
                self.log.warn(
                    f"Cannot write {output_loop['name']} for job {self.config['job_id']} to SSD"
                )
                return None
        df_transformed = self.create_output_loop(df, output_loop)
        df_transformed = self.collapse_stupid_lists(df_transformed, output_loop)
        df_transformed = self.add_metadata(df_transformed, output_loop)
        return df_transformed

    @exception_handler
    def load_data(self, df, output_loop):
        """Collect test_data locally and write to CSV.
        :param df: DataFrame to print.
        :return: None
        """
        loop_outpath = f"{self.config[Constant_Config.STEP_CONFIG]['output_loop_location']}/{output_loop['name']}"
        if len(output_loop["partition"]) > 0:
            df.printSchema()
            df.write.parquet(
                loop_outpath,
                mode="append",
                partitionBy=output_loop["partition"],
            )
        else:
            df.write.parquet(loop_outpath, mode="overwrite", compression="gzip")
        self.log.info(
            f"Succeeded writing json data for loop {output_loop['name']} to {loop_outpath}"
        )
        return None

    @exception_handler
    def run_etl(self):
        """
        Overloading run_etl to run transform & load on each separate output-loop/SSD table
        """
        for k, v in self.config[Constant_Config.STEP_CONFIG].items():
            self.log.info(f"{k}:")
            self.log.info(v)
        data = self.extract_data()
        for output_loop in self.config[Constant_Config.STEP_CONFIG]["outputs"]:
            data_transformed = self.transform_data(data, output_loop)
            if data_transformed is not None:
                self.load_data(data_transformed, output_loop)

        return


if __name__ == "__main__":
    run_json_to_pq = JSONToParquet()
    run_json_to_pq.main()



"""
test_etl_jobs.py
~~~~~~~~~~~~~~~
This module contains unit tests for the transformation steps of the ETL
job defined in etl_job_base.py. It makes use of a local version of PySpark
that is bundled with the PySpark package.
"""

import pytest
import pandas as pd
from tests.spark_context_fixture import spark_context, spark_session
from dependencies import aws, config
from constants.constants_configs import ProjectTest, Config as Constant_Config
import os

CONFIG_PROJECT_TEST_FILE_NAME = 'project_test.yaml'

# this to create fixture for spark_sessions
pytestmark = pytest.mark.usefixtures("spark_session")


def get_test_etl_job_parameters():
    '''

    The unittest configuration and jobs are placed in configs/project/project_test.yaml file
    This function get the list of unittest jobs from the config file (test_etl_job.jobs) and
    provide it as a fixture to the test_etl_jobs function. Pytest generates one unittest per
    job.
    :return:
    '''
    project_configs = config.load_framework_config()[Constant_Config.PROJECT_TEST]
    return project_configs[ProjectTest.TEST_ETL_JOB][ProjectTest.TEST_ETL_JOB_JOBS]


@pytest.mark.parametrize('etl_job_test_module', get_test_etl_job_parameters())
def test_etl_jobs(spark_session, etl_job_test_module):
    """
    :param spark_session:
    :param etl_job_test_module:
    :return:
    """

    # get configuration values (stored in configs/project/project_test.yaml)
    project_config = config.load_framework_config()
    project_test_configs = project_config[Constant_Config.PROJECT_TEST]
    storage_bucket = project_test_configs[ProjectTest.AWS_TEST_DIRECTORY][ProjectTest.AWS_TEST_DIRECTORY_BUCKET]
    storage_key = project_test_configs[ProjectTest.AWS_TEST_DIRECTORY][ProjectTest.AWS_TEST_DIRECTORY_BASE_KEY]
    # storage_aws_path = f"s3a://{storage_bucket}/{storage_key}"
    data_path = os.environ['DATA_PATH'] # Example: ~/cadmium/cadmium/tests/pickle/
    storage_aws_path= data_path

    # get the test module_name, class name, function and version(datetime)
    # from the etl_job_test_module fixture
    module_name = etl_job_test_module[0]
    class_name = etl_job_test_module[1]
    func_name = etl_job_test_module[2]
    version = etl_job_test_module[3]  # datetime
    etl_name = module_name.split(".")[-1]

    file_name_args = "{}test_data-{}-{}-args_{}.json".format(
        storage_key, etl_name, func_name, version)
    file_path_input = "{}test_data-{}-{}-input_{}.pickle".format(
        storage_aws_path, etl_name, func_name, version)
    file_path_output = "{}test_data-{}-{}-output_{}.pickle".format(
        storage_aws_path, etl_name, func_name, version)

    cadmium_aws = aws.cadmium_aws()
    args_dict = cadmium_aws.read_json_from_bucket(storage_bucket, file_name_args)
    # the input data for testing
    input_df = spark_session.createDataFrame(pd.read_pickle(file_path_input))
    # the expected data for testing
    expected_output = pd.read_pickle(file_path_output)

    module = __import__(module_name)  # import module
    my_class = getattr(getattr(module, "etl_job_base"), class_name)
    instance = my_class()  # init class
    instance.set_spark_session(spark_session)  # set class config
    instance.set_config(args_dict['config'])  # set class config

    instance.set_log()
    args_2_plus = args_dict['args_2_plus']
    func = getattr(instance, func_name)

    # this generates the current output for the test input data
    output_df = func(input_df, *args_2_plus).toPandas()

    # assert test using pandas assert
    pd.testing.assert_frame_equal(output_df, expected_output)



"""
spark.py
~~~~~~~~
Module containing helper function for use with Apache Spark
source: https://github.com/AlexIoannides/pyspark-example-project/blob/master/dependencies/spark.py
The original source for this file is above. It is heavily  modified for Cadmium use
"""

import __main__
from os import environ
from pyspark.sql import SparkSession


def start_spark(
    app_name="my_spark_app", master="local[*]", jar_packages=[], spark_config={}
):
    """Start Spark session, get Spark logger and load config files.
    Start a Spark session on the worker node and register the Spark
    application with the cluster. Note, that only the app_name argument
    will apply when this is called from a script sent to spark-submit.
    All other arguments exist solely for testing the script from within
    an interactive Python console.

    The function checks the enclosing environment to see if it is being
    run from inside an interactive console session or from an
    environment which has a `DEBUG` environment variable set (e.g.
    setting `DEBUG=1` as an environment variable as part of a debug
    configuration within an IDE such as Visual Studio Code or PyCharm.
    In this scenario, the function uses all available function arguments
    to start a PySpark driver from the local PySpark package as opposed
    to using the spark-submit and Spark cluster defaults. This will also
    use local module imports, as opposed to those in the zip archive
    sent to spark via the --py-files flag in spark-submit.

    :param app_name: Name of Spark app.
    :param master: Cluster connection details (defaults to local[*]).
    :param jar_packages: List of Spark JAR package names.
    :param files: List of files to send to Spark cluster (master and
        workers).
    :param spark_config: Dictionary of config key-value pairs.
    :return: A tuple of references to the Spark session, logger and
        config dict (only if available).
    """

    # detect execution environment
    flag_repl = not (hasattr(__main__, "__file__"))
    flag_debug = "DEBUG" in environ.keys()
    flag_repl = flag_debug = True

    if not (flag_repl or flag_debug):
        # get Spark session factory
        spark_builder = SparkSession.builder.appName(app_name)
    else:
        # get Spark session factory
        # spark_builder = SparkSession.builder.master(master).appName(app_name)
        spark_builder = SparkSession.builder.appName(app_name)

        # create Spark JAR packages string
        spark_jars_packages = ",".join(list(jar_packages))
        spark_builder.config("spark.jars.packages", spark_jars_packages)
        spark_builder.config("spark.sql.parquet.fs.optimized.committer.optimization-enabled", True)
        # add other config params
        for key, val in spark_config.items():
            spark_builder.config(key, val)

    # create session and retrieve Spark logger object
    spark_sess = spark_builder.getOrCreate()

    return spark_sess



from constants.constants_configs import Config as Constant_Config
from dependencies.exception import exception_handler
from jobs.flexi_etl import FlexiEtl
from utils.audit import Audit


class FlexiAudit(FlexiEtl):
    @exception_handler
    def unpivot_audit(self):
        # find  epi_member_id for claims
        audit = Audit(
            self.spark,
            self.config[Constant_Config.LOCATION_MASTER],
            self.get_config_for_step('audit_threshold'),
            self.log
        )

        df_unpivot_audit = audit.unpivot(self.spark.table("aggr_nitrogen_yearly_dashboard"))

        return df_unpivot_audit

    @staticmethod
    def print_df(df, load_settings, load_name):
        print_limit = load_settings.get('limit', 10)
        df.show(print_limit)
        if load_settings.get('count', False):
            print(f"count is {df.count()}")

    def load_data(self):
        self.log.info("Load...")
        step_config = self.config[Constant_Config.STEP_CONFIG]
        if step_config.get('Load', None) is None:
            return

        loads_containing_bad_measurements = []
        for df_name, df_settings in step_config['Load'].items():
            df: DataFrame = self.spark.table(df_name)
            self.log.info(f"key {df_name}")
            print(df_settings)

            if 'verify' in df_settings and df_settings['verify']:
                # Cache results as we may be accessing them multiple times to check whether or not the results are good
                df.cache()
                if df.count() > 0:
                    loads_containing_bad_measurements.append(df_name)

            for load_name, load_settings in df_settings['loads'].items():
                print(load_name)
                if "type" not in load_settings:
                    self.log.warn(f"load type has not been specified for table: {df_name}, load: {load_name}")
                    raise Exception(f"load type has not been specified for table: {df_name}, load: {load_name}")

                load_type = load_settings["type"]
                if load_type == "parquet":
                    self.save_to_parquet(df, load_settings, load_name)
                elif load_type == "postgres":
                    self.save_to_postgres(df, load_settings, load_name)
                elif load_type == "csv":
                    self.save_to_csv(df, load_settings, load_name)
                elif load_type == "print":
                    self.print_df(df, load_settings, load_name)

            df.unpersist()

        if len(loads_containing_bad_measurements) > 0:
            raise Exception(
                f"Found measurements outside expected thresholds. See the following loads: {loads_containing_bad_measurements}.")

        return None


if __name__ == "__main__":
    run_flexi_audit = FlexiAudit()
    run_flexi_audit.main()
